#include <ESP8266WiFi.h>
#include <EEPROM.h>
#include <ESP8266httpUpdate.h>

/** 
 *  flash: DOUT, 40Mhz, 80Mhz
    19.01.2020  V1.1    --: initial code
    27.01.2020  V1.2    --: DS18B20-code added
    03.02.2020  V1.3    bf: pick temp in web page not as IO_INT
    06.04.2020  V1.4    fr: save hourly values and show simple chart
    06.04.2020  V1.5    fr: load/save 24h-values
    09.04.2020  V1.6    --: save 24h values to EEPROM disabled
    09.04.2020  V1.6    fr: show last days and month summaries
    12.04.2020  V1.7    --: bugfixes, bigger chart
    12.04.2020  V1.8    fr: minutes watt values
    26.04.2020  V1.9    fr: more values on 30-day chart
    02.05.2020  V1.10   bf: neg minute values on 00:00
    02.05.2020  V1.10   fr: save to RTC-MEM every hour
    03.05.2020  V1.11   fr: load today kwh from RTC, temp warning >40deg
    03.05.2020  V1.12   fr: show 10sec-low-pass value
    18.06.2020  V1.13   bf: html-chart remove stripping factor to 20
    21.07.2020  V1.14   bf: char '!' allowed in ssid
    24.07.2020  V1.15   fr: allow run as access point
**/

extern "C" {
  #include "user_interface.h"
  #include "iob.h"
  #include "config.h"
}

#define LED   2

#define M_ESP_SETUP 0
#define M_ESP_RUN   1

int runmode=M_ESP_SETUP;   /* configure esp */

void _event_kwh( IobVar *v, int idx )
{
//  hlw_set_kWhtoday( v->u.f.val );
}

void _event_setup( IobVar *v, int idx )
{
  char *cmd=v->u.s.val;

  if( !strncmp(cmd,"UPDATE ",7) )
  {
    _doUpdate( cmd+7 );
  }
  else if ( !strcmp(cmd,"FACTORYRESET") )
  {
    ReconfMode();
    delay(100);
    writeRTC();
		memorizeSave();
    ESP.reset();
    delay(100);
  }
  else if ( !strcmp(cmd,"resetkWh") )
  {
    //acs712_reset_kWh();
  }
}

WiFiEventHandler gotIpEventHandler, disconnectedEventHandler;

static uint32_t boot_reason=0;

String boot_reason_text( void )
{
  switch( boot_reason )
  {
  case 0 : return (String)("[0] Power reboot");
  case 1 : return (String)("[1] Hardware WDT reset");
  case 2 : return (String)("[2] Fatal exception");
  case 3 : return (String)("[3] Software watchdog reset");
  case 4 : return (String)("[4] Software reset");
  case 5 : return (String)("[5] Deep-sleep");
  case 6 : return (String)("[6] Hardware reset");
  default : return (String)("[")+boot_reason+"] unknown";
  }
}

static int disco_count=0;

void setup()
{
  int i;
  struct rst_info *rtc_info = system_get_rst_info();
  boot_reason = rtc_info->reason;

#ifdef SERIAL_DEBUG
  Serial.begin(115200);
  Serial.printf("myResetInfo->reason %x \n", boot_reason);
#endif

  pinMode(0, INPUT_PULLUP );
  pinMode(LED, OUTPUT);
  digitalWrite(LED, 1); //Led port ausschalten

  ReadEEProm();

  // pinMode(pinbtn,INPUT_PULLUP);

  wifi_set_sleep_type(NONE_SLEEP_T);
  acs712_adc_Init();
  startDs18b20();

  if ( runmode != M_ESP_SETUP )
    initMemorize();

  if ( runmode == M_ESP_SETUP )
  {     
     prepareWiFiSetup(0);
     prepareWiFiConfSetup();

     prepareIobSetup();
     iobSetHandler( IDX_SETUP, _event_setup );
     iobSetHandler( IDX_KWH, _event_kwh );
     return;
  }

  gotIpEventHandler = WiFi.onStationModeGotIP([](const WiFiEventStationModeGotIP& event)
  {
#ifdef SERIAL_DEBUG
    Serial.print("Station connected, IP: ");
    Serial.println(WiFi.localIP());
#endif
  });

  disconnectedEventHandler = WiFi.onStationModeDisconnected([](const WiFiEventStationModeDisconnected& event)
  {
#ifdef SERIAL_DEBUG
    Serial.println("Station disconnected");
#endif
    disco_count++;
  });

  prepareWiFiConf();
  prepareWiFiConfSetup();

#ifdef SERIAL_DEBUG
  Serial.println("start in run mode");
  Serial.print(F("FlashSizebyChipID: "));
  Serial.println(ESP.getFlashChipSizeByChipId());
#endif
  prepareIobSetup();
  iobSetHandler( IDX_SETUP, _event_setup );
  iobSetHandler( IDX_KWH, _event_kwh );
}

void loop()
{
static int actloops=0;
static int obtn=0;
static uint32_t subloop=0;
static int  last_btn0=0;
static int  btn0_cnt=0;
static int switch_cfg=0;
static int zone=0;

#ifdef ENA_BTNPROG
  if ( !digitalRead(0) )
  {
    if ( last_btn0 )
      btn0_cnt++;

    if ( btn0_cnt == 800 )
      switch_cfg=1;

    last_btn0=1;
  }
  else
  {
    last_btn0=0;
    btn0_cnt=0;
  }
#endif

  if ( switch_cfg )
  {
    switch_cfg++;
    zone=switch_cfg / 50;
    digitalWrite(LED,(zone&1)?0:1);

    if ( switch_cfg > 1500 )
    {
      ReconfMode();
      delay(100);
      writeRTC();
      memorizeSave();
      ESP.reset();
      delay(100);
    }
  }

  if ( !( subloop % 5 ) )
  {
    runWiFiConfLoop();

    if ( runmode != M_ESP_SETUP )
      WiFiConnectLoop();

    timerLoop();	/* read time via ntp */
    runIobLoop( runmode == M_ESP_SETUP );

    ds18b20_Loop();
  }

  acs712_adc_loop();

  subloop++;

  delay(2);
}
