
#include <ESP8266WiFi.h>
#include <SyncClient.h>
#include "iob.h"

extern "C" {
  #include "user_interface.h"
}

#define PP_HTONL(x) ((((x) & 0xff) << 24) | \
                     (((x) & 0xff00) << 8) | \
                     (((x) & 0xff0000UL) >> 8) | \
(((x) & 0xff000000UL) >> 24))
#define ntohl(a) PP_HTONL(a)
#define htonl(a) PP_HTONL(a)

class IobQueue
{
public:
  WiFiClient        client;
  int               step;
  int               wrong_cnt;
  int               used;
  int               watch_time;
  uint8_t           data[512];
  uint16_t          start;
  uint16_t          fill;
  uint8_t           set_used;
};

class RemIobQueue
{
public:
  SyncClient        client;
  int               step;
  int               wrong_cnt;
  int               used;
  int               watch_time;
  uint8_t           data[512];
  uint16_t          start;
  uint16_t          fill;
  uint8_t           set_used;
  uint32_t          last_rcvd_time;
  uint32_t          last_wd_time;
  byte              inv_wd_cnt;
};

#define IOBPORT       9773

#undef SERIAL_DEBUG

extern int runmode;   /* configure esp */

WiFiServer iob(IOBPORT);

static IobQueue      iobclist[10];
static RemIobQueue   remiob;

#define IO_INT             1
#define IO_FLOAT           2
#define IO_STRING          3

#define IOB_ACCESS         1
#define IOB_NOACCESS       2
#define IOB_CHANGE_INT     3
#define IOB_CHANGE_FLOAT   4
#define IOB_CHANGE_STRING  5
#define IOB_STATE          6
#define IOB_CLEAR_RESPONSE 9

#define IOB_MASK_ACCESS    (1L<<0)
#define IOB_MASK_CHANGE    (1L<<1)
#define IOB_MASK_ICHANGE   (1L<<2)
#define IOB_MASK_FCHANGE   (1L<<3)
#define IOB_MASK_SCHANGE   (1L<<4)
#define IOB_MASK_STATE     (1L<<5)

#define IF_BACKSEND        (1L<<19)
#define IF_RDONLY          (1L<<21)
#define IF_BINARY          (1L<<20)
#define IF_PULSE           (1L<<18)

typedef struct _Evr
{
  int8_t      cn;  /* aka line */
  uint32_t    mask;
  uint32_t    param;
  struct _Evr *next;
} Evr;

static IobVar iobvars[] = {
  { "/version", IO_STRING, "",   IF_RDONLY,      0,  0 },
  { "/setup",   IO_STRING, "",   0,              0,  1 },
  { "/current", IO_FLOAT,  "",   0,              0,  1 },
  { "/power",   IO_FLOAT,  "",   0,              0,  1 },
  { "/totalkWh",IO_FLOAT,  "",   0,              0,  1 },
  { "/totalWmin",IO_FLOAT, "",   0,              0,  1 },
  { "/voltage", IO_FLOAT,  "",   0,              0,  1 },
  { "/temp",    IO_FLOAT,  "",   0,              0,  1 },
#define NUMVARS 8
};


static void iobSetHandler( int idx, IHProc proc )
{
  if (( idx < 0 ) || ( idx >= NUMVARS ))
    return;
  iobvars[idx].handler = proc;
}

static void _doAccessPV( int cn, int ac, char **av )
{
  uint32_t  mask, ll;
  char      *pvname=av[0];
  int       i;

  mask=_HexString(av[1]);
  ll=_HexString(av[2]);

  for( i=0; i< NUMVARS; i++ )
  {
    if ( !strcmp(pvname,iobvars[i].pvname) )
       break;
  }
  if ( i == NUMVARS )
  {
    uint8_t noacdata[10]={ 4, 0, 0, IOB_NOACCESS, 0x2d, 0x26, 0, 0, 0, 0 };
    memcpy(noacdata+6,&ll,4);
    iobclist[cn].client.write( (const uint8_t*)noacdata, 10 );
    return;
  }
  int off=1;
  char data[512];
  uint32_t  ull[16];
  ull[0]     = ll;
  ull[1] = i+1;
  ull[2] = htonl(iobvars[i].type);
  ull[3] = htonl(iobvars[i].state);
  ull[4] = 0;    /* state1 */
  ull[5] = 0;    /* state2 */
  ull[6] = 0;    /* l_chg.tv_sec */
  ull[7] = 0;    /* l_chg.tv_usec */

  int sz = 8*4; /* hdr */

  switch( iobvars[i].type )
  {
  case IO_INT :
    ull[8] = htonl(iobvars[i].u.i.val);
    ull[9] = htonl(iobvars[i].u.i.rangeb);
    ull[10] = htonl(iobvars[i].u.i.rangee);
    sz += 12;
    memcpy(data+6,ull,sz);
    strcpy(data+6+sz,iobvars[i].unit);
    sz += strlen(iobvars[i].unit)+1;
    break;
  case IO_FLOAT :
    memcpy(ull+10,&iobvars[i].u.f.val,8);
    memcpy(ull+12,&iobvars[i].u.f.rangeb,8);
    memcpy(ull+14,&iobvars[i].u.f.rangee,8);
    {
      int x;
      for( x=8; x<14; x++ )
        ull[x] = htonl(ull[x+2+!(x%2)]);
    }
    sz += 24;
    memcpy(data+6,ull,sz);
    strcpy(data+6+sz,iobvars[i].unit);
    sz += strlen(iobvars[i].unit)+1;
    break;
  case IO_STRING :
    memcpy(data+6,ull,sz);
    strcpy(data+6+sz,iobvars[i].u.s.val);
    sz += strlen(iobvars[i].u.s.val)+1;
    break;
  }

  data[0] = sz & 0xff;
  data[1] = (sz>>8) & 0xff;
  data[2] = 0;
  data[3] = IOB_ACCESS;
  data[4] = 0x2d;
  data[5] = 0x26;
  iobclist[cn].client.write( (const uint8_t*)data, sz+6 );

  Evr *evr;
  
  for( evr = iobvars[i].evr; evr; evr=evr->next )
  {
    if (( evr->param == ll ) && ( evr->cn == cn ))
    {
      evr->mask = mask | IOB_MASK_ACCESS;
      break;
    }
  }
  
  if ( !evr )
  {
    evr=(Evr*)malloc(sizeof(Evr));
    memset(evr,0,sizeof(Evr));
    evr->mask = mask | IOB_MASK_ACCESS;
    evr->param = ll;
    evr->cn = cn;

    evr->next = iobvars[i].evr;
    iobvars[i].evr = evr;
  }
}

static void _sendEvent( int cn, int idx, uint32_t mask, uint32_t type, int source )
{
  uint32_t  ll[10], bsend=0;
  char      data[512];
  char      *payload=data+9;
  char      *bcopyto;
  Evr       *evr;
  int sz=0;

  if ( config.use_remiob && (source!=SRC_RIOB) &&
       ( iobvars[idx].access_mode & 2 ))   /* access okay */
  {
    switch( iobvars[idx].type )
    {
    case IO_INT :
      payload[0]=0xC9;
      payload[1]=sizeof(long);
      memcpy(payload+2,&iobvars[idx].srvid,sizeof(long));
      payload[6]=0xC5;
      payload[7]=sizeof(long);
      memcpy(ll,&iobvars[idx].u.i.val,sizeof(long));
      ll[0]=htonl(ll[0]);
      memcpy(payload+8,ll,sizeof(long));
      bsend=12;
      break;
    case IO_FLOAT :
      payload[0]=0xC9;
      payload[1]=sizeof(long);
      memcpy(payload+2,&iobvars[idx].srvid,sizeof(long));
      payload[6]=0xC6;
      payload[7]=sizeof(long)*2;
      memcpy(ll,&iobvars[idx].u.f.val,sizeof(long)*2);
      ll[1]=htonl(ll[1]);
      ll[2]=htonl(ll[0]);
      memcpy(payload+8,ll+1,sizeof(long)*2);
      bsend=16;
      break;
    case IO_STRING :
#ifdef SERIAL_DEBUG
Serial.println("build STRING payload");
#endif
      payload[0]=0xC9;
      payload[1]=sizeof(long);
      memcpy(payload+2,&iobvars[idx].srvid,sizeof(long));
      payload[6]=0xC7;
      bsend=strlen(iobvars[idx].u.s.val);
      if ( bsend )
      {
        if ( bsend > 255 )
            bsend=255;
        memcpy(payload+8,iobvars[idx].u.s.val,bsend);
        *(payload+8+bsend)=0;
        bsend++;
        payload[7]=bsend;
        bsend+=8;
      }
      else
      {
        payload[7]=1;
        payload[8]=0;
        bsend=9;
      }
      break;
    }
    bsend+=3;
    data[0] = bsend & 0xff;
    data[1] = (bsend >> 8) & 0xff;
    data[2] = 0;
    data[3] = 2;
    data[4] = 0x2d;
    data[5] = 0x26;
    data[6] = 'X';
    data[7] = 'X';
    data[8] = 0;
#ifdef SERIAL_DEBUG
Serial.println("send2 remote iob");
#endif
    if ( remiob.used )
      remiob.client.write( (const uint8_t*)data, bsend+6 );
  }

  if (( source == SRC_IOB )  || /* do not send back own event */
      !config.use_iob )
    return;

  switch( type )
  {
  case IOB_CHANGE_INT :
    ll[0]=0;
    ll[1]=htonl( iobvars[idx].u.i.val );
    ll[2]=0;  /* tv-sec */
    ll[3]=0;
    ll[4]=htonl( iobvars[idx].state );
    ll[5]=0;  /* state1 */
    ll[6]=0;
    sz=4*7;
    break;
  case IOB_CHANGE_FLOAT :
    ll[0]=0;
    memcpy(ll+3,&iobvars[idx].u.f.val,8 );
    ll[1]=htonl(ll[4]);
    ll[2]=htonl(ll[3]);
    ll[3]=0;  /* tv-sec */
    ll[4]=0;
    ll[5]=htonl( iobvars[idx].state );
    ll[6]=0;  /* state1 */
    ll[7]=0;
    sz=4*8;
    break;
  case IOB_CHANGE_STRING :
    ll[0]=0;
    ll[1]=0;  /* tv-sec */
    ll[2]=0;
    sz=4*3;
  }
  if ( !sz )
    return;
  memcpy(data+6,(char*)ll, sz);
  if ( type == IOB_CHANGE_STRING )
  {
    int slen=strlen(iobvars[idx].u.s.val);
    memcpy(data+6+sz,iobvars[idx].u.s.val,slen+1);
    sz+=slen+1;
  }
  bcopyto=data+6+sz;
  sz += 4;

  for( evr = iobvars[idx].evr; evr; evr=evr->next )
  {
    if ( evr->cn == -1 )
      continue;
    if ( !( evr->mask & mask ) )
      continue;
    if ( ! iobclist[ evr->cn ].used )
      continue;
    ll[0] = evr->param;
    memcpy(data+6,(char*)ll,4);
    bsend = (cn == evr->cn) ? htonl(IF_BACKSEND) : 0;
    memcpy(bcopyto,(char*)&bsend,4);

    data[0] = sz & 0xff;
    data[1] = (sz>>8) & 0xff;
    data[2] = 0;
    data[3] = type;
    data[4] = 0x2d;
    data[5] = 0x26;
    iobclist[evr->cn].client.write( (const uint8_t*)data, sz+6 );
  }
}

#define NOPULSE 0
#define ISPULSE 1
#define NOCHGCHECK  0
#define CHGCHECK    1

static void _setVar( int cn, int idx, char *value, int8_t ispulse, uint8_t check_chg, int source )
{
  uint8_t valchg=1;
  uint32_t mask=0;
  uint32_t type=0;
  int      changed=0;

#if 0
#ifdef SERIAL_DEBUG
  Serial.printf("_setVar start: cn=%d, idx=%d, val=%s, source=%d\n",cn,idx,value,source);
#endif
#endif
  if ( iobvars[idx].state & IF_RDONLY )
    return;
  if ((( iobvars[idx].state & IF_PULSE ) ? 1 : 0 ) ^ ispulse )
  {
    if ( ispulse )
      iobvars[idx].state |= IF_PULSE;
    else
      iobvars[idx].state &= ~IF_PULSE;
  }

  if ( iobvars[idx].type == IO_FLOAT )
  {
    double ofval = iobvars[idx].u.f.val;
    iobvars[idx].u.f.val = atof(value);
    if ( ofval != iobvars[idx].u.f.val )
      changed=1;
    if ( check_chg && !changed )
      return;
    mask = IOB_MASK_CHANGE | IOB_MASK_FCHANGE;
    type = IOB_CHANGE_FLOAT;
  }
  else if ( iobvars[idx].type == IO_INT )
  {
    int32_t oival = iobvars[idx].u.i.val;
    iobvars[idx].u.i.val = atoi(value);
    if ( oival != iobvars[idx].u.i.val )
      changed=1;
    if ( check_chg && !changed )
      return;
    mask = IOB_MASK_CHANGE | IOB_MASK_ICHANGE;
    type = IOB_CHANGE_INT;
  }
  else if ( iobvars[idx].type == IO_STRING )
  {
    char *osval = iobvars[idx].u.s.val;
    iobvars[idx].u.s.val = strdup(value);
    if ( strcmp(osval,iobvars[idx].u.s.val))
      changed=1;
    if ( check_chg && !changed )
    {
      free( osval );
      return;
    }
    free( osval );
    mask = IOB_MASK_CHANGE | IOB_MASK_SCHANGE;
    type = IOB_CHANGE_STRING;
  }
  _sendEvent( cn, idx, mask, type, source );
#if 0
#ifdef SERIAL_DEBUG
  Serial.printf("_setVar: chg=%d, cn=%d, hnd=%p, source=%d\n",changed,cn,iobvars[idx].handler,source);
#endif
#endif
  if ( changed && ( cn != -1 ) && iobvars[idx].handler )
    iobvars[idx].handler( iobvars+idx, idx );
  else if ( changed && ( cn == -1 ) && iobvars[idx].handler && ( source != SRC_BTN ))
    iobvars[idx].handler( iobvars+idx, idx );
}

static void _setFloatVar( int cn, int idx,  double value, int8_t ispulse, uint8_t check_chg, int source )
{
  uint8_t  valchg=1;
  uint32_t mask=0;
  uint32_t type=0;
  int      changed=0;

#if 0
#ifdef SERIAL_DEBUG
  Serial.printf("_setFloatVar start: cn=%d, idx=%d, val=%d.%02d, source=%d\n",cn,idx,(int)value,(int)(value*100)%100,source);
#endif
#endif
  if ( iobvars[idx].state & IF_RDONLY )
    return;
  if ((( iobvars[idx].state & IF_PULSE ) ? 1 : 0 ) ^ ispulse )
  {
    if ( ispulse )
      iobvars[idx].state |= IF_PULSE;
    else
      iobvars[idx].state &= ~IF_PULSE;
  }

  if ( iobvars[idx].type == IO_FLOAT )
  {
    double ofval = iobvars[idx].u.f.val;
    iobvars[idx].u.f.val = value;
    if ( ofval != iobvars[idx].u.f.val )
      changed=1;
    if ( check_chg && !changed )
      return;
    mask = IOB_MASK_CHANGE | IOB_MASK_FCHANGE;
    type = IOB_CHANGE_FLOAT;
  }
  else if ( iobvars[idx].type == IO_INT )
  {
    int32_t oival = iobvars[idx].u.i.val;
    iobvars[idx].u.i.val = (int)value;
    if ( oival != iobvars[idx].u.i.val )
      changed=1;
    if ( check_chg && !changed )
      return;
    mask = IOB_MASK_CHANGE | IOB_MASK_ICHANGE;
    type = IOB_CHANGE_INT;
  }
  else if ( iobvars[idx].type == IO_STRING )
  {
    char txtval[32];
    char *osval = iobvars[idx].u.s.val;
    sprintf(txtval,"%d.%03d",(int)value,(int)(value*1000)%1000);
    iobvars[idx].u.s.val = strdup(txtval);
    if ( strcmp(osval,iobvars[idx].u.s.val))
      changed=1;
    if ( check_chg && !changed )
    {
      free( osval );
      return;
    }
    free( osval );
    mask = IOB_MASK_CHANGE | IOB_MASK_SCHANGE;
    type = IOB_CHANGE_STRING;
  }
  _sendEvent( cn, idx, mask, type, source );
#ifdef SERIAL_DEBUG
  Serial.printf("_setVar: chg=%d, cn=%d, hnd=%p, source=%d\n",changed,cn,iobvars[idx].handler,source);
#endif
  if ( changed && ( cn != -1 ) && iobvars[idx].handler )
    iobvars[idx].handler( iobvars+idx, idx );
  else if ( changed && ( cn == -1 ) && iobvars[idx].handler && ( source != SRC_BTN ))
    iobvars[idx].handler( iobvars+idx, idx );
}

static void _setIntVar( int cn, int idx, int value, int8_t ispulse, uint8_t check_chg, int source )
{
  uint8_t  valchg=1;
  uint32_t mask=0;
  uint32_t type=0;
  int      changed=0;

#ifdef SERIAL_DEBUG
  Serial.printf("_setIntVar start: cn=%d, idx=%d, val=%d, source=%d\n",cn,idx,(int)value,source);
#endif
  if ( iobvars[idx].state & IF_RDONLY )
    return;
  if ((( iobvars[idx].state & IF_PULSE ) ? 1 : 0 ) ^ ispulse )
  {
    if ( ispulse )
      iobvars[idx].state |= IF_PULSE;
    else
      iobvars[idx].state &= ~IF_PULSE;
  }

  if ( iobvars[idx].type == IO_FLOAT )
  {
    double ofval = iobvars[idx].u.f.val;
    iobvars[idx].u.f.val = (double)value;
    if ( ofval != iobvars[idx].u.f.val )
      changed=1;
    if ( check_chg && !changed )
      return;
    mask = IOB_MASK_CHANGE | IOB_MASK_FCHANGE;
    type = IOB_CHANGE_FLOAT;
  }
  else if ( iobvars[idx].type == IO_INT )
  {
    int32_t oival = iobvars[idx].u.i.val;
    iobvars[idx].u.i.val = value;
    if ( oival != iobvars[idx].u.i.val )
      changed=1;
    if ( check_chg && !changed )
      return;
    mask = IOB_MASK_CHANGE | IOB_MASK_ICHANGE;
    type = IOB_CHANGE_INT;
  }
  else if ( iobvars[idx].type == IO_STRING )
  {
    char txtval[32];
    char *osval = iobvars[idx].u.s.val;
    sprintf(txtval,"%d",value);
    iobvars[idx].u.s.val = strdup(txtval);
    if ( strcmp(osval,iobvars[idx].u.s.val))
      changed=1;
    if ( check_chg && !changed )
    {
      free( osval );
      return;
    }
    free( osval );
    mask = IOB_MASK_CHANGE | IOB_MASK_SCHANGE;
    type = IOB_CHANGE_STRING;
  }
  _sendEvent( cn, idx, mask, type, source );
#ifdef SERIAL_DEBUG
  Serial.printf("_setVar: chg=%d, cn=%d, hnd=%p, source=%d\n",changed,cn,iobvars[idx].handler,source);
#endif
  if ( changed && ( cn != -1 ) && iobvars[idx].handler )
    iobvars[idx].handler( iobvars+idx, idx );
  else if ( changed && ( cn == -1 ) && iobvars[idx].handler && ( source != SRC_BTN ))
    iobvars[idx].handler( iobvars+idx, idx );
}

static void _doSetValue( int cn, int ac, char **av, uint8_t check_chg )
{
  iobclist[cn].set_used=1;
  int   idx = _HexString(av[0])-1;

  if (( idx < 0 ) || ( idx >= NUMVARS ))
    return;
  _setVar( cn, idx, av[1], NOPULSE, check_chg, SRC_IOB );
}

static void _doSetNamedValue( int cn, int ac, char **av, uint8_t check_chg )
{
  char  *pvname=av[0];
  int   i;
  iobclist[cn].set_used=1;
  for( i=0; i< NUMVARS; i++ )
  {
    if ( !strcmp(pvname,iobvars[i].pvname) )
       break;
  }
  if ( i==NUMVARS)
    return;
  _setVar( cn, i, av[1], NOPULSE, check_chg, SRC_IOB );
}

typedef struct _PulseTimer
{
  int                 cn;
  int                 idx;
  uint32_t            milli_start;
  uint32_t            milli_diff;
  char                *backval;
  struct _PulseTimer  *next;
} PulseTimer;

static PulseTimer  *pulse_root=0;

static void _AddPulseTimer( int cn, int idx, uint32_t diff, char *backval )
{
  PulseTimer *pt=(PulseTimer*)malloc(sizeof(PulseTimer));
  memset(pt,0,sizeof(PulseTimer));
  pt->cn=cn;
  pt->idx=idx;
  pt->milli_diff=diff;
  pt->milli_start = millis() | 0x80000000;
  pt->backval = strdup(backval);

  pt->next = pulse_root;
  pulse_root = pt;
}

static void _runPulseTimer( void )
{
  PulseTimer *pt, *pp, *pn;
  uint32_t  now=millis() | 0x80000000;
  uint32_t  diff;

  for( pt=pulse_root, pp=0; pt; pt=pn )
  {
    pn=pt->next;

    pt->milli_start |= 0x80000000;
    
    if ( now < pt->milli_start )
    {
      pt->milli_start &= 0x7fffffff;    /* remove bit31 */
      pt->milli_start |= 1;             /* make != 0 */
    }
    diff=now-pt->milli_start;
    if ( diff < pt->milli_diff )
    {
      pp=pt;
      continue;
    }
    _setVar( pt->cn, pt->idx, pt->backval, ISPULSE, CHGCHECK, SRC_IOB );
    free( pt->backval );
    if ( pp )
      pp->next = pn;
    else
      pulse_root = pn;
    free( pt );
  }
}

static void _doSetPulse( int cn, int ac, char **av )
{
  iobclist[cn].set_used=1;
  int   idx = _HexString(av[0])-1;

  if (( idx < 0 ) || ( idx >= NUMVARS ))
    return;
  _setVar( cn, idx, av[1], ISPULSE, CHGCHECK, SRC_IOB );
  _AddPulseTimer( cn, idx, atoi(av[2]), av[3] );
}

static void _doSetNamedPulse( int cn, int ac, char **av )
{
  char  *pvname=av[0];
  int   i;
  iobclist[cn].set_used=1;
  for( i=0; i< NUMVARS; i++ )
  {
    if ( !strcmp(pvname,iobvars[i].pvname) )
       break;
  }
  if ( i==NUMVARS)
    return;
  _setVar( cn, i, av[1], ISPULSE, CHGCHECK, SRC_IOB );
  _AddPulseTimer( cn, i, atoi(av[2]), av[3] );
}

static void _clearAccess( int cn, int idx )
{
  Evr   *evr, *evp, *evn;

  if (( idx < 0 ) || ( idx >= NUMVARS ))
    return;
  for( evp=0, evr=iobvars[idx].evr; evr; evr=evn )
  {
    evn=evr->next;
    if ( evr->cn == cn )
    {
      uint8_t cadata[10]={ 4, 0, 0, IOB_CLEAR_RESPONSE, 0x2d, 0x26, 0, 0, 0, 0 };
      memcpy(cadata+6,&evr->param,4);
      iobclist[cn].client.write( (const uint8_t*)cadata, 10 );
      if( evp )
        evp->next = evn;
      else
        iobvars[idx].evr = evn;
      free(evr);
      continue;
    }
    evp=evr;
  }
}

static void _doClearAccess( int cn, int ac, char **av )
{
  Evr   *evr, *evp, *evn;
  int   idx = _HexString(av[0])-1;

  _clearAccess( cn, idx );
}

static void _doClearNamedAccess( int cn, int ac, char **av )
{
  char  *pvname=av[0];
  int   i;

  for( i=0; i< NUMVARS; i++ )
  {
    if ( !strcmp(pvname,iobvars[i].pvname) )
       break;
  }
  if ( i==NUMVARS)
    return;
  _clearAccess( cn, i );
}

static int getAsInt( int intype, int ival, double fval, char *sval )
{
  if ( intype == IO_INT )
    return ival;
  if ( intype == IO_FLOAT )
    return fval;
  if ( intype == IO_STRING )
    return (sval ? atoi(sval) : 0);
  return 0;
}

int iobGetIntVal( int idx )
{
  if (( idx < 0 ) || ( idx >= NUMVARS ))
     return 0;
  switch( iobvars[idx].type )
  {
    case IO_INT :
      return iobvars[idx].u.i.val;
    case IO_FLOAT :
      return (int)iobvars[idx].u.f.val;
    case IO_STRING :
      return atoi(iobvars[idx].u.s.val);
  }
  return 0;
}

static double getAsFloat( int intype, int ival, double fval, char *sval )
{
  if ( intype == IO_INT )
    return ival;
  if ( intype == IO_FLOAT )
    return fval;
  if ( intype == IO_STRING )
    return (sval ? atof(sval) : 0);
  return 0;
}

static char *getAsString( int intype, int ival, double fval, char *sval, char *tmp )
{
  if ( intype == IO_INT )
  {
    sprintf(tmp,"%d",ival);
    return tmp;
  }
  if ( intype == IO_FLOAT )
  {
    sprintf(tmp,"%d.%03d",(int)fval,(int)(fval*1000)%1000);
    return tmp;
  }
  if ( intype == IO_STRING )
    return sval;
  *tmp=0;
  return tmp;
}

static void doSetBinVal( int cn, char *data, int len )
{
  char      *p, *p2, *np;
  uint32_t  sl;
  int       idx=-1;
  uint8_t   nocheck=0;
  uint32_t  ll[2];
  double    fval;
  int32_t   ival;
  char      *sval=0;
  int       intype=0;
  uint32_t  mask, type;
  int       changed=0;

  for( p=data; p<data+len; p=np )
  {
    sl=p[1];
    p2=p+2;
    if ( !sl )
    {
      memcpy(&sl,p+2,4);
      sl=ntohl(sl);
      p2 = p+6;
      np = p + sl + 6;
    }
    else
      np = p + sl + 2;

    switch( p[0] )
    {
    case 0xC9 :   /* pv-ptr */
       memcpy(&idx,p2,4);
       idx -= 1;
       break;
    case 0xC8 :   /* pv-name */
       for( idx=0; idx< NUMVARS; idx++ )
       {
         if ( !strcmp(p2,iobvars[idx].pvname) )
           break;
       }
       break;
    case 0xC4 :  /* NOCHECK */
       nocheck = p2[0];
       break;
    case 0xC5 :  /* INT-VAL */
       intype = IO_INT;
       memcpy(&ival,p2,4);
       ival = ntohl(ival);
       break;
    case 0xC6 :  /* FLOAT-VAL */
       intype = IO_FLOAT;
       memcpy(ll,p2,8);
       if ( htonl(1) != 1 )
       {
         uint32_t  my_long = htonl(ll[0]);
         ll[0] = ll[1];
         ll[1] = my_long;
       }
       memcpy(&fval,ll,8);
       break;
    case 0xC7 : /* STRG */
       intype = IO_STRING;
       sval = (char*)malloc(sl+1);
       memset(sval,0,sl+1);
       memcpy(sval,p2,sl);
       break;
    }
  }
  while( 1 )
  {
    if (( idx < 0 ) || ( idx >= NUMVARS ))
      break;

    if ( iobvars[idx].state & IF_RDONLY )
      break;

    changed=0;
    if ( iobvars[idx].type == IO_INT )
    {
      int32_t oival = iobvars[idx].u.i.val;
      iobvars[idx].u.i.val = getAsInt( intype, ival, fval, sval );
      if ( oival != iobvars[idx].u.i.val )
        changed=1;
      if ( !nocheck && !changed )
        break;
      mask = IOB_MASK_CHANGE | IOB_MASK_ICHANGE;
      type = IOB_CHANGE_INT;
    }
    else if( iobvars[idx].type == IO_FLOAT )
    {
      double ofval = iobvars[idx].u.f.val;
      iobvars[idx].u.f.val = getAsFloat( intype, ival, fval, sval );
      if ( ofval != iobvars[idx].u.f.val )
        changed=1;
      if ( !nocheck && !changed )
        break;
      mask = IOB_MASK_CHANGE | IOB_MASK_FCHANGE;
      type = IOB_CHANGE_FLOAT;
    }
    else if ( iobvars[idx].type == IO_STRING )
    {
      char tmp[32];
      char *osval = iobvars[idx].u.s.val;
      char *sval = getAsString( intype, ival, fval, sval, tmp );
      iobvars[idx].u.s.val = strdup(sval);
      if ( strcmp(osval,sval) )
        changed=1;
      if ( !nocheck && !changed )
      {
        free( osval );
        break;
      }
      mask = IOB_MASK_CHANGE | IOB_MASK_SCHANGE;
      type = IOB_CHANGE_STRING;
    }
    _sendEvent( cn, idx, mask, type, 0 );
    if ( changed && ( cn != -1 ) && iobvars[idx].handler )
      iobvars[idx].handler( iobvars+idx, idx );
    break;
  }
  if ( sval )
    free( sval );
}

static void doPacket( int i, unsigned char *p, int len )
{
  unsigned char *data=p+6;
  int port = (p[5]<<8)+p[4];
  uint8_t ptype=p[3];
  char          *params=0;
  char          **av=0;
  int           ac;

  if ( port == 0xfefe )
  {
    iobclist[i].step=99;
    return;
  }
  if ( port == 0xf3f3 )
  {
    const uint8_t locdata[10]={ 0, 0, 0, 0, 0xf3, 0xf3 };
    uint32_t  wt;
    memcpy(&wt,(char*)p+6,sizeof(int32_t));
    wt=ntohl(wt);
    iobclist[i].watch_time=wt;

    iobclist[i].client.write( locdata, 6 );  /* answer */
    return;
  }
  /* assemble packet here */
  if ( p[3] == 0x02 ) /* set bin val */
  {
    doSetBinVal( i, (char*)p+6, len );
    return;
  }
  if ( p[3] != 0 )
    return;
  params=strchr((char*)p+6,' ');
  
  if ( params )
  {
    *params=0;
    params++;
    
    av=VfkCutStrg( params, &ac, ',' );
  }
  if ( !strcmp((char*)p+6,"AC") )
  {
    _doAccessPV( i, ac, av );
  }
  else if ( !strcmp((char*)p+6,"SV") || !strcmp((char*)p+6,"USV"))
  {
    _doSetValue( i, ac, av, CHGCHECK );
  }
  else if ( !strcmp((char*)p+6,"S-V") || !strcmp((char*)p+6,"US-V"))
  {
    _doSetValue( i, ac, av, NOCHGCHECK );
  }
  else if ( !strcmp((char*)p+6,"SN") )
  {
    _doSetNamedValue( i, ac, av, CHGCHECK );
  }
  else if ( !strcmp((char*)p+6,"PV") || !strcmp((char*)p+6,"UPV"))
  {
    _doSetPulse( i, ac, av );
  }
  else if ( !strcmp((char*)p+6,"PN") || !strcmp((char*)p+6,"UPN"))
  {
    _doSetNamedPulse( i, ac, av );
  }
  else if ( !strcmp((char*)p+6,"CA") )
  {
    _doClearAccess( i, ac, av );
  }
  else if ( !strcmp((char*)p+6,"CNA") )
  {
    _doClearNamedAccess( i, ac, av );
  }
  if ( av )
    free(av);
}

static void doRemiobPacket( unsigned char *p, int len )
{
  unsigned char *data=p+6;
  int port = (p[5]<<8)+p[4];
  uint8_t ptype=p[3];
  uint32_t  idx;
#if 0
#ifdef SERIAL_DEBUG
  Serial.println(F("assemble packet now...."));
  {
    int i;
    uint8_t *up = p;
    for( i=0;i<len+6;i++,up++ )
      Serial.printf("%02x ",*up);
    Serial.println("");
  }
#endif
#endif

  remiob.last_rcvd_time = millis() | 0x80000000;

  if ( port == 0xfefe )
  {
    remiob.client.stop();
		remiob.used=0;
    return;
  }
  if ( port == 0xf3f3 )
  {
    const uint8_t locdata[6]={ 0, 0, 0, 0, 0xf3, 0xf3 };

#ifdef SERIAL_DEBUG
Serial.println("watchdog received");
#endif
    remiob.client.write( locdata, 6 );

    return;
  }
  /* assemble packet here */
  switch( ptype )
  {
  case IOB_CHANGE_FLOAT :
     memcpy(&idx,data,4);
     if ( idx < NUMVARS )
     {
      long ll[10];
      double fval;
      char value[64];
      long back=0;
#ifdef SERIAL_DEBUG
      Serial.printf("IOB_CHANGE_FLOAT : %s\n",iobvars[idx].pvname);
#endif
      memcpy(ll,data,len>40?40:len);
      ll[0]=ntohl(2);
      ll[1]=ntohl(ll[1]);
      memcpy(&fval,ll,sizeof(long)*2);
      sprintf(value,"%d.%03d",(int)fval,(int)(fval*1000)%1000);
      if ( len > sizeof(long)*8 )
        back=ntohl(ll[8]);
      if ( !( back & IF_BACKSEND ) )
        _iobSetVar( idx, value, SRC_RIOB );
     }
     break;
  case IOB_CHANGE_STRING :
     memcpy(&idx,data,4);
     if ( idx < NUMVARS )
     {
      long ll[10];
      double fval;
      char *value;
      long back=0;
      int  sl;
#ifdef SERIAL_DEBUG
      Serial.printf("IOB_CHANGE_STRING : %s\n",iobvars[idx].pvname);
#endif

      value=(char*)data+sizeof(long)*3;
      len -= sizeof(long)*3;
      sl=strlen(value);
      if ( len > sl+1 )
      {
        memcpy(ll+4,value+sl+1,sizeof(long));
        back=ntohl(ll[4]);
      }
      if ( !( back & IF_BACKSEND ) )
        _iobSetVar( idx, sl?value:(char*)"", SRC_RIOB );
     }
     break;
  case IOB_CHANGE_INT :
     memcpy(&idx,data,4);
     if ( idx < NUMVARS )
     {
      long ll[10];
      char value[64];
      long back=0;
      memcpy(ll,data,len>40?40:len);
      sprintf(value,"%d",ntohl(ll[1]));
      if ( len > sizeof(long)*7 )
        back=ntohl(ll[7]);
#ifdef SERIAL_DEBUG
      Serial.printf("IOB_CHANGE_INT : %s\n",iobvars[idx].pvname);
#endif
      if ( !( back & IF_BACKSEND ) )
        _iobSetVar( idx, value, SRC_RIOB );
     }
     break;
  case IOB_NOACCESS :
     memcpy(&idx,data,4);
     if ( idx >= NUMVARS )
        return;
#ifdef SERIAL_DEBUG
     Serial.printf("IOB_NOACCESS : %s\n",iobvars[idx].pvname);
#endif
     iobvars[idx].access_mode &= ~2;  /* remove bit 1 */
     break;
  case IOB_ACCESS :
     memcpy(&idx,data,4);
     if ( idx < NUMVARS )
     {
      unsigned long srvid;
      unsigned long pvtype;
      long          ival;
      double        fval;
      char          *sval=0;
#ifdef SERIAL_DEBUG
     Serial.printf("IOB_ACCESS : %s\n",iobvars[idx].pvname);
#endif
      memcpy(&srvid,data+sizeof(long),sizeof(long));
      memcpy(&pvtype,data+sizeof(long)*2,sizeof(long));
      iobvars[idx].srvid=srvid;
      pvtype=ntohl(pvtype);
      iobvars[idx].access_mode |= 2;  /* set bit 1 */
      _provideVar( idx );
      switch( pvtype )
      {
      case IO_INT :
        memcpy(&ival,data+sizeof(long)*8,sizeof(long));
				ival=ntohl(ival);
				if ( idx != IDX_KWH )
					_iobSetIntVar( idx, ival, SRC_RIOB );
        else if (ival>iobvars[idx].u.i.val)
        {
          double fval = ival+iobvars[idx].u.i.val;
          _iobSetFloatVar( idx, fval, SRC_RIOB );
        }
        break;
      case IO_FLOAT :
        {
          long ll[3];
          memcpy(ll+1,data+sizeof(long)*8,sizeof(long)*2);
          ll[0]=ntohl(ll[2]);
          ll[1]=ntohl(ll[1]);
          memcpy(&fval,ll,sizeof(long)*2);
        }
				if ( idx != IDX_KWH )
					_iobSetFloatVar( idx, fval, SRC_RIOB );
        else if (fval>iobvars[idx].u.f.val)
          _iobSetFloatVar( idx, fval+iobvars[idx].u.f.val, SRC_RIOB );
        break;
      case IO_STRING :
        ival=strlen((char*)data+sizeof(long)*8);
        sval=(char*)malloc(ival+1);
        memcpy(sval,data+sizeof(long)*8,ival+1);
				if ( idx != IDX_KWH )
					_iobSetVar( idx, sval, SRC_RIOB );
        free(sval);
        break;
      }
     }
     break;
   default:
     break;
  }
}

static void DoQueuedIobClient( int i )
{
  if ( !iobclist[i].client )
  {
    iobclist[i].step=99;
    return;
  }
  if ( iobclist[i].step == 0 )
  {
    if(iobclist[i].client.connected())
        iobclist[i].step=1;
    else
    {
      iobclist[i].wrong_cnt++;
      if ( iobclist[i].wrong_cnt > 100 )
      {
        iobclist[i].step=99;
        return;
      }
    }
  }

  WiFiClient client = iobclist[i].client;
  
  if( !client.connected() )
  {
    iobclist[i].step=99;
    return;
  }
  if ( iobclist[i].step == 1 )		/* read requests */
  {
    int b;
    int fill = iobclist[i].fill;

    if ( client.available() > 0 )
    {
      uint8_t *p=iobclist[i].data + fill;
      size_t avail=client.available();

      if ( avail > 512 - fill )
         avail = 512 - fill;
      if ( client.readBytes(p,avail) != -1 )
        fill+=avail;
    }
    iobclist[i].fill = fill;
    int avail= iobclist[i].fill - iobclist[i].start;

    while ( avail > 5 )  /* skit-header okay */
    {
      unsigned char *p=iobclist[i].data + iobclist[i].start;
      int pcklen= (p[1]<<8) + p[0];
      if ( pcklen+6 > avail )
         break;
      doPacket( i, p, pcklen );
      pcklen += 6;
      iobclist[i].start += pcklen;
      avail -= pcklen;
    }
    if ( iobclist[i].start == iobclist[i].fill )
    {
      iobclist[i].start=0;
      iobclist[i].fill=0;
    }
    else
    {
      memcpy(iobclist[i].data,iobclist[i].data + iobclist[i].start, iobclist[i].fill - iobclist[i].start);
      iobclist[i].fill -= iobclist[i].start;
      iobclist[i].start = 0;
    }
  }
}

static void _iobSetVar( int idx, char *value, int source )
{
  _setVar( -1, idx, value, NOPULSE, CHGCHECK, source );
}

static void _iobSetFloatVar( int idx, double value, int source )
{
  _setFloatVar( -1, idx, value, NOPULSE, CHGCHECK, source );
}

static void _iobSetIntVar( int idx, int value, int source )
{
  _setIntVar( -1, idx, value, NOPULSE, CHGCHECK, source );
}

static void _iobPulseIntVar( int idx, int value, int source, int diff, int back )
{
  char bval[32];
  sprintf(bval,"%d",back);
  _setIntVar( -1, idx, value, ISPULSE, CHGCHECK, source );
  _AddPulseTimer( -1, idx, diff, bval );
}

static void prepareIobSetup( void )
{
  char data[256];
  char ext[16];
#ifdef SERIAL_DEBUG
  if ( config.use_iob )
     Serial.println(F("prepare as iob"));
#endif
  int   i;

  for( i=0; i<NUMVARS; i++ )
  {
    switch( iobvars[i].type )
    {
    case IO_INT :
      iobvars[i].u.i.rangeb = 0;
      iobvars[i].u.i.rangee = 1;
      iobvars[i].u.i.val = 0;
      break;
    case IO_FLOAT :
      iobvars[i].u.f.rangeb = 0;
      iobvars[i].u.f.rangee = 100;
      iobvars[i].u.f.val = 0;
      break;
    case IO_STRING :
      iobvars[i].u.s.val = 0;
      break;
    }
  }
  strcpy(data,VERSION_STRG);

  iobvars[IDX_VERSION].u.s.val = strdup(data);

  for( i=0; i<NUMVARS; i++ )
  {
    if (( iobvars[i].type == IO_STRING ) && !iobvars[i].u.s.val )
      iobvars[i].u.s.val = strdup("");
  }

  if ( config.use_iob )
	  iob.begin();
}

static unsigned long last_access_time=0x80000000;
unsigned long last_ac_diff;

#define PVOFFS  MODUL

int num_send_access=0;
int num_ac_loop=0;
int num_acv_called=0;
static void _accessVars( void )
{
  int   i, sz, nac=0;
  char  data[64];
static int time_pause=5000;
  int   todo=0;
  
  unsigned long now = millis() | 0x80000000;
  num_acv_called++;

  if ( last_access_time > now )
		last_access_time &= 0x7fffffff;
  last_ac_diff = now - last_access_time;
  last_access_time |= 0x80000000;
  if ( last_ac_diff < time_pause )
     return;
  last_access_time = now;
  num_ac_loop++;
  for( i=0; i<NUMVARS; i++ )
  {
    if ( !( iobvars[i].access_mode & 1 ) )  /* need access ? */
      continue;
    if ( iobvars[i].access_mode & 2 )       /* already accessed ? */
      continue;
    if ( iobvars[i].access_mode & 4 )       /* already made */
      continue;
    todo++;
  }
  if ( !todo )
  {
    for( i=0; i<NUMVARS; i++ )
      iobvars[i].access_mode &= 0xffB;     /* 1111 1111 1011 */
  }
  for( i=0; i<NUMVARS; i++ )
  {
    if ( !( iobvars[i].access_mode & 1 ) )  /* need access ? */
      continue;
    if ( iobvars[i].access_mode & 2 )       /* already accessed ? */
      continue;
    if ( iobvars[i].access_mode & 4 )       /* already made */
      continue;
    iobvars[i].access_mode |= 4;
/* access /Sonoff/name/switch */
    sprintf(data+6,"AC /" PVOFFS "/%s%s,%lx,%lx",config.DevName.c_str(),iobvars[i].pvname,IOB_MASK_ACCESS|IOB_MASK_CHANGE,i);
    sz=strlen(data+6)+1;
    data[0] = sz & 0xff;
    data[1] = (sz>>8) & 0xff;
    data[2] = 0;
    data[3] = 0;
    data[4] = 0x2d;
    data[5] = 0x26;
    remiob.client.write( (const uint8_t*)data, sz+6 );
    num_send_access++;
    nac++;
    if ( nac==2 )
      break;
  }
  if ( nac == 2 )
    time_pause=2000;
  else
    time_pause=5000;
}

int num_provide=0;
static void _provideVar( int idx )
{
  char  data[64];
  int   sz;
  
  sprintf(data+6,"UD %lx,1",iobvars[idx].srvid);
  sz=strlen(data+6)+1;
  data[0] = sz & 0xff;
  data[1] = (sz>>8) & 0xff;
  data[2] = 0;
  data[3] = 0;
  data[4] = 0x2d;
  data[5] = 0x26;
  remiob.client.write( (const uint8_t*)data, sz+6 );
  num_provide++;
}

void iobSetRemNamed( char *pvname, char *strg )
{
  char data[128];
  int  sz;
  if ( !remiob.used )
    return;
  if ( strlen(strg)+strlen(pvname)> 96 )
    return;
  sprintf(data+6,"SN /" PVOFFS "/%s%s,'%s'",config.DevName.c_str(),pvname,strg);
  sz=strlen(data+6)+1;
  data[0] = sz & 0xff;
  data[1] = (sz>>8) & 0xff;
  data[2] = 0;
  data[3] = 0;
  data[4] = 0x2d;
  data[5] = 0x26;
  if ( remiob.used && remiob.client.connected() )
    remiob.client.write( (const uint8_t*)data, sz+6 );
}

static int num_asy_connect=0;
static int num_asy_done=0;
extern int mxmode;
static int retry_speed=5000;  /* 5 Sekunden */

static void _makeConnectedRemiob( void )
{
static unsigned long lasttry=0x80000000;
       unsigned long now=millis() | 0x80000000;
       unsigned long lastdiff;
static uint32_t pid=1;

	if ( remiob.used && !remiob.client.connected() )
	{
#ifdef SERIAL_DEBUG
  Serial.println("remote iob lost");
#endif
// MX1
    mxmode=1;
		remiob.client.stop();
		remiob.used=0;
    num_asy_connect=0;
	}
	if ( remiob.used && remiob.client.connected() )
		return;
  if ( lasttry > now )
		lasttry &= 0x7fffffff;
	lastdiff = now - lasttry;
  lasttry |= 0x80000000;
	if ( !remiob.used && (lastdiff >= retry_speed) )
	{
	  lasttry = now;
#ifdef SERIAL_DEBUG
    Serial.println("try connect");
#endif
    num_asy_connect++;
    if ( num_asy_connect == 6 )  /* ab3.Versuch 5 Sekunden */
    {
	    memorizeSave();
      ESP.reset();
      delay(100);      
    }
		if ( remiob.client.asyncConnect( toIPAddress(config.iobIP), 9773 ) )
		{
      char data[64];
      int sz, i;
#ifdef SERIAL_DEBUG
      Serial.println("remiob connected");
#endif
      mxmode=0;
      retry_speed=2000;
			remiob.used=1;
      remiob.fill=0;
      remiob.start=0;
      remiob.last_rcvd_time = millis() | 0x80000000;
/* setup watch time 15secs */
      memset(data,0,10);
      data[0] = 4;
      data[4] = 0xf3;
      data[5] = 0xf3;
      data[9] = 15;   /* watch timer */
      remiob.client.write( (const uint8_t*)data, 10 );
/* send line info */
      sprintf(data+6,"SA 'sonoff' %d '%s' '-'",pid++,config.DevName.c_str());
      sz=strlen(data+6)+1;
      data[0] = sz & 0xff;
      data[1] = (sz>>8) & 0xff;
      data[2] = 0;
      data[3] = 0;
      data[4] = 0x2d;
      data[5] = 0x26;
      remiob.client.write( (const uint8_t*)data, sz+6 );
/* set named /Sonoff/name/setup */
      sprintf(data+6,"SN /" PVOFFS "/%s%s,''",config.DevName.c_str(),iobvars[IDX_SETUP].pvname);
      sz=strlen(data+6)+1;
      data[0] = sz & 0xff;
      data[1] = (sz>>8) & 0xff;
      data[2] = 0;
      data[3] = 0;
      data[4] = 0x2d;
      data[5] = 0x26;
      remiob.client.write( (const uint8_t*)data, sz+6 );
/* set named /Sonoff/name/version */
      sprintf(data+6,"SN /" PVOFFS "/%s/version,'%s'",config.DevName.c_str(),iobvars[IDX_VERSION].u.s.val);
      sz=strlen(data+6)+1;
      data[0] = sz & 0xff;
      data[1] = (sz>>8) & 0xff;
      data[2] = 0;
      data[3] = 0;
      data[4] = 0x2d;
      data[5] = 0x26;
      remiob.client.write( (const uint8_t*)data, sz+6 );

      num_asy_done++;
      last_access_time=0x80000000;
      for( i=0; i< NUMVARS; i++ )
        iobvars[ i ].access_mode &= ~2;
 		}
	}
}

void iobDropLines( void )
{
  int i, clnum=0;
	if ( remiob.used )
  {
		remiob.client.stop();
    remiob.used=0;
  }
	for( i=0; i<10; i++ )
	{
		if ( !iobclist[i].used )
			continue;
		if ( iobclist[i].step != 98 )
    {
      clnum++;
		  iobclist[i].step = 99;
    }
	}
  if ( clnum )
		runIobLoop( 1 );
}

static void _SendRemoteWatchdog( void )
{
  uint32_t      now = millis() | 0x80000000;
  int           diff;
  char          data[10];

  if ( !config.use_remiob || !w_connected )
    return;

  if ( !remiob.last_rcvd_time )
    return;

  if ( !remiob.used && !remiob.client.connected() )
    return;

  if ( now < remiob.last_rcvd_time )
  {
    remiob.last_rcvd_time &= 0x7fffffff;
    remiob.last_rcvd_time |= 1;
  }
  diff = now - remiob.last_rcvd_time;
  remiob.last_rcvd_time |= 0x80000000;

  if ( diff > 15000 )
  {
#ifdef SERIAL_DEBUG
Serial.println("-connection lost-");
#endif
    remiob.client.stop();
    remiob.start=0;
    remiob.fill=0;
    remiob.used=0;
    remiob.last_rcvd_time=0;
    mxmode=1;
    num_asy_connect=0;
  }
}

static void _checkRemoteIncoming( void )
{
  int		b, rc;
  int		fill;
  size_t	avail;
  uint8_t	*p;

  if ( !w_connected )
    return;
  if ( !config.use_remiob )
  {
    return;
  }
  wifi_set_sleep_type(NONE_SLEEP_T);
	_makeConnectedRemiob();

	if ( !remiob.used && !remiob.client.connected() || (remiob.client.available()<1) )
		return;

  fill = remiob.fill;
  p=remiob.data + fill;
  avail=remiob.client.available();
  if ( avail > 512 - fill )
    avail=512-fill;
  rc= remiob.client.readBytes(p,avail);
  if ( rc != -1 )
    fill+=rc;
 
  remiob.fill = fill;
  avail= remiob.fill - remiob.start;

  while ( avail > 5 )  /* skit-header okay */
  {
    unsigned char *p=remiob.data + remiob.start;
    int pcklen= (p[1]<<8) + p[0];
    if ( pcklen+6 > avail )
       break;
    doRemiobPacket( p, pcklen );
    pcklen += 6;
    remiob.start += pcklen;
    avail = remiob.fill - remiob.start;
  }
  if ( remiob.start == remiob.fill )
  {
    remiob.start=0;
    remiob.fill=0;
  }
  else
  {
    memcpy(remiob.data,remiob.data + remiob.start, remiob.fill - remiob.start);
    remiob.fill -= remiob.start;
    remiob.start = 0;
  }
  if ( remiob.used && remiob.client.connected() )
    _accessVars();
}

static void runIobLoop( int configmode )
{
  int       i;
  int       clnum=0;
  WiFiClient client;

#ifdef SERIAL_DEBUG
Serial.println("runIobLoop");
#endif

  if ( config.use_iob )
  {
		client = iob.available();

		while ( client )
		{
			for( i=0;i<10;i++)
			{
				if ( !iobclist[i].used )
					break;
			}
			if ( i==10 )
			{
				client.stop();
#ifdef SERIAL_DEBUG
				Serial.println("**** too many clients ****");
#endif
				break;
			}
			iobclist[i].client=client;
			iobclist[i].step=0;   /* 0: read from client */
			iobclist[i].used=1;
			iobclist[i].wrong_cnt=0;
			break;
		}
		clnum=0;
		for( i=0; i<10; i++ )
		{
			if ( !iobclist[i].used )
				continue;
			clnum++;
			if ( iobclist[i].step < 90 )
				DoQueuedIobClient( i );
			if ( iobclist[i].step == 99 ) /* remove from queue */
			{
	/* remove client depend mem */
				Evr *evr, *evp, *evn;
				int x;
				for( x=0; x<NUMVARS; x++ )
				{
					for( evp=0, evr=iobvars[x].evr; evr; evr=evn )
					{
						evn=evr->next;
						if ( evr->cn != i )
						{
							evp=evr;
							continue; 
						}
						if ( evp )
							evp->next = evn;
						else
							iobvars[x].evr = evn;
						free( evr );
					}
				}
				PulseTimer *pt;
				for( pt=pulse_root; pt; pt=pt->next )
				{
					if ( pt->cn == i )
						pt->cn = -1;
				}
				iobclist[i].step=98;
				delay(10);
			}
			else if ( iobclist[i].step == 98 ) /* remove from queue */
			{
				if ( iobclist[i].client )
					iobclist[i].client.stop();
				iobclist[i].used=0;
			}
		}
    _runPulseTimer();
  }
  if ( !configmode )
  {
	_checkRemoteIncoming();
    _SendRemoteWatchdog();
  }
#ifdef SERIAL_DEBUG
Serial.println("runIobLoop done");
#endif
}

String iobDebug( int full )
{
  String res;
  int    tinv= 0;
  
  if ( iobvars[IDX_TEMP].u.f.val > 44.99 )
    tinv=2;
  else if ( iobvars[IDX_TEMP].u.f.val > 39.99 )
    tinv=1;

  res=F("<BR><BR><table cellpadding=\"3\"><tr><td>Current</td><td>");
  res+=iobvars[IDX_CURRENT].u.f.val;
  res+=F(" A</td></tr><tr><td>Power</td><td>");
  res+=iobvars[IDX_POWER].u.f.val;
  res+=F(" W</td></tr><tr><td>Voltage</td><td>");
  res+=iobvars[IDX_VOLTAGE].u.f.val;
  res+=F(" mV</td></tr><tr><td>total</td><td>");
  res+=iobvars[IDX_WM].u.f.val;
  res+=F(" Wmin</td></tr><tr><td>total</td><td>");
  res+=iobvars[IDX_KWH].u.f.val;
  res+=F(" kWh</td></tr><tr><td>Temperature</td><td>");
  if ( tinv == 2 )
	  res+=F("<span style=\"background-color: red\">");
  if ( tinv == 1 )
    res+=F("<span style=\"background-color: orange\">");
  res+=iobvars[IDX_TEMP].u.f.val;
  res+=F(" &#176;C");
  if ( tinv )
	  res+=F("</span>");
  res+=F("</td></tr></table>");

  if ( config.use_remiob )
  {
    if ( full )
    {
      res+=F("<br>remiob.used : ");
      res+=remiob.used ? F("yes") : F("no");
      res+=F("<br>remiob.connected : ");
      res+=remiob.client.connected() ? F("yes") : F("no");
      res+=F("<br>remiob.available : ");
      res+=remiob.client.available();
      res+=F("<br>num_asy_connect : ");
      res+=num_asy_connect;
      res+=F("<br>num_asy_done : ");
      res+=num_asy_done;
      res+=F("<br>num_provide : ");
      res+=num_provide;
      res+=F("<br>num_ac_loop : ");
      res+=num_ac_loop;
      res+=F("<br>num_send_access : ");
      res+=num_send_access;
      res+=F("<br>num_acv_called : ");
      res+=num_acv_called;
      res+=F("<br>last_ac_diff : ");
      res+=last_ac_diff;
    }
  }
  return res;
}
