
#include <ESP8266WiFi.h>
//#include <TimeLib.h>
#include <Timezone.h>


IPAddress timeServerIP(195,43,74,123);
const char *ntpServerName = "0.europe.pool.ntp.org";
#define NTP_PACKET_SIZE 48

static byte packetBuffer[ NTP_PACKET_SIZE ];
WiFiUDP udp;

//Central European Time (Frankfurt, Paris)
TimeChangeRule CEST = {"CEST", Last, Sun, Mar, 2, 120};    //Central European Summer Time
TimeChangeRule CET = {"CET", Last, Sun, Oct, 3, 60};       //Central European Standard Time
Timezone CE(CEST, CET);

static uint32_t stime=0;
static uint32_t num_ntp_sent=0;

static struct
{
   byte     req_pending;
   uint32_t req_millis;
   uint32_t millis;
   time_t   stamp;
   time_t   secs;
   byte     initialized;
   uint32_t wrong;
   uint32_t num_ntp_recvd;
   uint32_t num_ntp_ok;
//   int      auf;    /* minuten bis SA */
//   int      unter;    /* minuten bis SU */
   uint32_t num_ntp_timeout;
   time_t   no1time;	/* 1st received time ever */
} itime;

time_t getCurrentSecs( void )
{
	if ( !itime.stamp )
		return 0;
	return itime.secs;
}

// send an NTP request to the time server at the given address
void sendNTPpacket( void )
{
#ifdef SERIAL_DEBUG
  Serial.println("sending NTP packet...");
  Serial.print("ntpServerName:");
  Serial.println(ntpServerName);
  Serial.print("timeServerIP:");
  Serial.println(timeServerIP);
#endif
#ifdef IOBLOG
  char txt[96];
  sprintf(txt,"sendNTPpacket : itime.secs=%ld",itime.secs);
  iobSetRemNamed("/log",txt);
#endif

  itime.req_pending=1;
  itime.req_millis=millis() | 0x80000000;
  if (( timeServerIP[0] == 0 ) || (itime.num_ntp_timeout > 10 ))
    WiFi.hostByName(ntpServerName,timeServerIP);
  if ( timeServerIP[0] == 0 )
    return;
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;

  num_ntp_sent++;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  udp.beginPacket(timeServerIP, 123); //NTP requests are to port 123
  udp.write(packetBuffer, NTP_PACKET_SIZE);
  udp.endPacket();
}

#ifdef SERIAL_DEBUG
//Function to print time with time zone
void printTime(time_t t, char *tz, char *loc)
{
  sPrintI00(hour(t));
  sPrintDigits(minute(t));
  sPrintDigits(second(t));
  Serial.print(' ');
  Serial.print(dayShortStr(weekday(t)));
  Serial.print(' ');
  sPrintI00(day(t));
  Serial.print(' ');
  Serial.print(monthShortStr(month(t)));
  Serial.print(' ');
  Serial.print(year(t));
  Serial.print(' ');
  Serial.print(tz);
  Serial.print(' ');
  Serial.print(loc);
  Serial.println();
}

//Print an integer in "00" format (with leading zero).
//Input value assumed to be between 0 and 99.
void sPrintI00(int val)
{
  if (val < 10) Serial.print('0');
  Serial.print(val, DEC);
  return;
}

//Print an integer in ":00" format (with leading zero).
//Input value assumed to be between 0 and 99.
void sPrintDigits(int val)
{
  Serial.print(':');
  if (val < 10) Serial.print('0');
  Serial.print(val, DEC);
}
#endif

void timerInitialize( void )
{
  memset(&itime,0,sizeof(itime));

  if ( !w_connected )
    return;

  WiFi.hostByName(ntpServerName,timeServerIP);

  udp.begin(2390);  /* localport */
  sendNTPpacket();
  itime.millis=itime.req_millis;
  itime.initialized=88; /* magic */
}

String timerDebug( int full )
{
  char txt[128];
  String result;

  result=F("<BR>itime.secs (now) : ");
  sprintf(txt,"%02d.%02d.%04d %02d:%02d:%02d",
         day(itime.secs),
         month(itime.secs),
         year(itime.secs),
         hour(itime.secs),
         minute(itime.secs),
         second(itime.secs));
  result+=txt;
  if(full)
  {
    sprintf(txt," (%d)",itime.secs);
    result+=txt;
  }
  result+=F("<BR>itime.stamp (last ntp) : ");
  sprintf(txt,"%02d.%02d.%04d %02d:%02d:%02d",
         day(itime.stamp),
         month(itime.stamp),
         year(itime.stamp),
         hour(itime.stamp),
         minute(itime.stamp),
         second(itime.stamp));
  result+=txt;
  if(full)
  {
    sprintf(txt," (%d)",itime.stamp);
    result+=txt;
  }
  result+=F("<BR>itime.no1time (booted) : ");
  sprintf(txt,"%02d.%02d.%04d %02d:%02d:%02d",
         day(itime.no1time),
         month(itime.no1time),
         year(itime.no1time),
         hour(itime.no1time),
         minute(itime.no1time),
         second(itime.no1time));
  result+=txt;
  if(full)
  {
    sprintf(txt," (%d)",itime.no1time);
    result+=txt;
  }
  result+=F("<BR>itime.req_pending : ");
  sprintf(txt,"%s",itime.req_pending ? "yes":"no");
  result+=txt;
  if(itime.wrong)
  {
    result+=F("<BR>itime.wrong : ");
    result+=itime.wrong;
  }
  result+=F("<BR>ntp-server-name : ");
  result+=ntpServerName;
  result+=F("<BR>ntp-server-ip : ");
  result+=timeServerIP.toString();
  if(full)
  {
    result+=F("<BR>num_ntp_sent : ");
    result+=num_ntp_sent;
    result+=F("<BR>itime.num_ntp_recvd : ");
    result+=itime.num_ntp_recvd;
    result+=F("<BR>itime.num_ntp_ok : ");
    result+=itime.num_ntp_ok;
    result+=F("<BR>itime.num_ntp_timeout : ");
    result+=itime.num_ntp_timeout;
  }
  return result;
}

void timerLoop( void )
{
  uint32_t  now, stime;
  uint32_t  diff;

  if ( itime.req_pending )  /* weve sent a Request */
  {
    int x=udp.parsePacket();
    if ( x )
    {
#ifdef SERIAL_DEBUG
Serial.print("udp.parsePacket() : ");
Serial.println(x);
#endif
      memset(packetBuffer, 0, NTP_PACKET_SIZE);
      udp.read(packetBuffer, NTP_PACKET_SIZE );
      if ( x < 44 )
        return;

#ifdef IOBLOG
      char txt[96];
      sprintf(txt,"parseUdpPacket : itime.secs=%ld, x=%d",itime.secs,x);
      iobSetRemNamed("/log",txt);
#endif
      itime.num_ntp_recvd++;

      unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
      unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);

      unsigned long secsSince1900 = highWord << 16 | lowWord;

#ifdef SERIAL_DEBUG
      Serial.print("Seconds since Jan 1 1900 = " );
      Serial.println(secsSince1900);

      Serial.print("Unix time = ");
#endif
      // Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
      const unsigned long seventyYears = 2208988800UL;
      // subtract seventy years:
      unsigned long epoch = secsSince1900 - seventyYears;
      // print Unix time:
#if 0
epoch=1542315258;   /* 21:54:18 (15.11.2018) */
epoch=1542257658;   /* 05:54:18 (15.11.2018) */
epoch=1542250818;   /* 04:00:54 */
#endif
      Serial.println(epoch);

      TimeChangeRule *tcr;
      time_t utc;
      utc = epoch;
      time_t new_stamp;

      if ( epoch < 1542315258 )  /* kann nicht sein , älter als 15.11.2018 ist nicht möglich */
      {
        itime.wrong++;
        return;
      }
      if ( itime.secs > 1542315258 )   /* this is a follow call */
      {
        if (( year(epoch) > year(itime.secs)+1 ) ||
            ( year(epoch) < year(itime.secs) ) )
        {
          itime.wrong++;
          return;
        }
      }

      new_stamp=CE.toLocal(utc,&tcr); /* stamp at itime.millis */
#ifdef IOBLOG
      sprintf(txt,"utc: %d, itime.stamp=%d, itime.secs",utc,itime.stamp,itime.secs);
      iobSetRemNamed("/log",txt);
      sprintf(txt,"old-ntp: %02d.%02d.%04d %02d:%02d:%02d",
         day(itime.stamp),
         month(itime.stamp),
         year(itime.stamp),
         hour(itime.stamp),
         minute(itime.stamp),
         second(itime.stamp));
      iobSetRemNamed("/log",txt);
      sprintf(txt,"new-ntp: %02d.%02d.%04d %02d:%02d:%02d",
         day(new_stamp),
         month(new_stamp),
         year(new_stamp),
         hour(new_stamp),
         minute(new_stamp),
         second(new_stamp));
      iobSetRemNamed("/log",txt);
#endif
      if ( new_stamp <= itime.stamp )
      {
        itime.wrong++;
        return;
      }

      stime=0;  /* mark as done */
      itime.millis = millis() | 0x80000000;

      itime.stamp = new_stamp;
      if ( !itime.no1time )
        itime.no1time=itime.stamp;
      itime.secs=itime.stamp;
#if 0
      itime.zerostamp = itime.secs-
	( hour(new_stamp)*3600+minute(new_stamp)*60+second(new_stamp) );
#endif
      itime.req_pending=0;
      itime.num_ntp_ok++;

#if 0
      if ( config.use_dg )
        CalcSRD( config.dg_breite, config.dg_laenge, 0, 1, &itime.auf, &itime.unter );
      else
        CalcSRD( 51.47888, 7.755, 0, 1, &itime.auf, &itime.unter );
#endif

#ifdef IOBLOG
      sprintf(txt,"ntp: %s %02d.%02d.%04d %02d:%02d:%02d",
        tcr -> abbrev,
        day(itime.stamp),
        month(itime.stamp),
        year(itime.stamp),
        hour(itime.stamp),
        minute(itime.stamp),
        second(itime.stamp));
      iobSetRemNamed("/log",txt);
#endif

#ifdef SERIAL_DEBUG
#ifndef IOBLOG
      char txt[96];
#endif
      printTime(utc, "UTC", "Universal Coordinated Time");
      printTime(itime.stamp, tcr -> abbrev, "Berlin");
      sprintf(txt,"ntp: %s %02d.%02d.%04d %02d:%02d:%02d",
        tcr -> abbrev,
        day(itime.stamp),
        month(itime.stamp),
        year(itime.stamp),
        hour(itime.stamp),
        minute(itime.stamp),
        second(itime.stamp));
      Serial.println(txt);
#endif
    }
  }
  if ( itime.initialized != 88 )
  {
    timerInitialize();
    return;
  }
  now = millis() | 0x80000000;
  stime = itime.millis | 0x80000000;
  if ( now < stime )
  {
    stime &= 0x7fffffff;    /* remove bit31 */
    stime |= 1;             /* make != 0 */
  }
  diff=now-stime;
  diff /= 1000;     /* wieviele ganze sekunden ? */
  if ( diff > 0 )
  {
    itime.millis += diff*1000;
    itime.millis |= 0x80000000;
    itime.secs += diff;

    if ( !itime.req_pending )
    {
#ifdef IOBLOG
      char txt[96];
      sprintf(txt,"timer : ntp-diff : %ld secs",itime.secs - itime.stamp);
      iobSetRemNamed("/log",txt);
#endif
      /* Jeden Tag nach 04:00 (wenn der letzte Versuch 43200 Sekunden her ist (1/2 Tag) , 10800=3h */
      /* 25 = every 25 hours */
      if ( ( hour(itime.secs) >= 4 ) && ( itime.secs - itime.stamp > 25*3600 ))
      {
	WiFi.hostByName(ntpServerName,timeServerIP);
#ifdef IOBLOG
	char txt[96];
	sprintf(txt,"send ntp-request now...");
	iobSetRemNamed("/log",txt);
#endif
	sendNTPpacket();
      }
    }
  }
  if ( itime.req_pending )
  {
    stime=itime.req_millis | 0x80000000;
    if ( now < stime )
    {
      stime &= 0x7fffffff;    /* remove bit31 */
      stime |= 1;             /* make != 0 */
    }
    diff=now-stime;
    if ( diff > 10000 )   /* 10 seconds */
    {
#ifdef SERIAL_DEBUG
      Serial.println("resend ntp request");
#endif
      itime.num_ntp_timeout++;
      sendNTPpacket();
    }
  }
}

#if 0
/************ Berechnung Sonnenaufgang/Sonnenuntergang ************/

double pi2=6.283185307179586476925286766559;
double pi=3.1415926535897932384626433832795;
double RAD = 0.017453292519943295769236907684886;

static double JulianischesDatum( int Jahr, int Monat, int Tag,
             int Stunde, int Minuten, double Sekunden )
{
    int   Gregor;
    if (Monat<=2)
    {
        Monat=Monat +12;
        Jahr = Jahr -1;
    }
    Gregor = (Jahr/400)-(Jahr/100)+(Jahr/4);  /* Gregorianischer Kalender */
    return 2400000.5+365.0*Jahr - 679004.0 + Gregor
          + (int)(30.6001*(Monat+1)) + Tag + Stunde/24.0
          + Minuten/1440.0 + Sekunden/86400.0;
}
static double InPi(double x)
{
    int n = (int)(x/pi2);
    x = x - n*pi2;
    if (x<0) x +=pi2;
    return x;
}

static double eps(double T) /* Neigung der Erdachse */
{
    return RAD*(23.43929111 + (-46.8150*T - 0.00059*T*T + 0.001813*T*T*T)/3600.0);
}

static double BerechneZeitgleichung(double *DK,double T)
{

    double RA_Mittel = 18.71506921
        + 2400.0513369*T +(2.5862e-5 - 1.72e-9*T)*T*T;

    double M  = InPi(pi2 * (0.993133 + 99.997361*T));
    double L  = InPi(pi2 * (  0.7859453 + M/pi2
            + (6893.0*sin(M)+72.0*sin(2.0*M)+6191.2*T) / 1296.0e3));
    double e = eps(T);
    double RA = atan(tan(L)*cos(e));
    double dRA;

    if (RA<0.0) RA+=pi;
    if (L>pi) RA+=pi;

    RA = 24.0*RA/pi2;

    *DK = asin(sin(e)*sin(L));

    /* Damit 0<=RA_Mittel<24 */
    RA_Mittel = 24.0*InPi(pi2*RA_Mittel/24.0)/pi2;

    dRA = RA_Mittel - RA;
    if (dRA < -12.0) dRA+=24.0;
    if (dRA > 12.0) dRA-=24.0;

    dRA = dRA* 1.0027379;

    return dRA ;
}

static void CalcSRD( double geoB, double geoL, int doff, int gmtplus,
        int *aufm, int *unterm )
{
    double JD2000 = 2451545.0;
    double DK;
    double h = -50.0/60.0*RAD;
    double B=geoB*RAD;
    double JD;
    double T;
    time_t  now;
    double Zeitgleichung;
    double Zeitdifferenz;
    double AufgangOrtszeit;
    double UntergangOrtszeit;
    double AufgangWeltzeit;
    double UntergangWeltzeit;
    double Aufgang;
    double Untergang;

    now=itime.secs+(doff*86400);

    JD = JulianischesDatum(year(itime.secs),
                month(itime.secs),
                day(itime.secs),
                12,0,0); /* Testdatum */

    if ( CE.locIsDST(itime.secs) )
        gmtplus++;

    T = ((JD+doff) - JD2000)/36525.0;

    Zeitgleichung = BerechneZeitgleichung(&DK,T);
    Zeitdifferenz = 12.0*acos((sin(h) - sin(B)*sin(DK)) / (cos(B)*cos(DK)))/pi;
    AufgangOrtszeit = 12.0 - Zeitdifferenz - Zeitgleichung;
    UntergangOrtszeit = 12.0 + Zeitdifferenz - Zeitgleichung;
    AufgangWeltzeit = AufgangOrtszeit - geoL /15.0;
    UntergangWeltzeit = UntergangOrtszeit - geoL /15.0;

    Aufgang = AufgangWeltzeit + gmtplus;         /* In Stunden */
    if (Aufgang<0.0) Aufgang +=24.0;
    else if (Aufgang>=24.0) Aufgang -=24.0;

    Untergang = UntergangWeltzeit + gmtplus;
    if (Untergang<0.0) Untergang +=24.0;
    else if (Untergang>=24.0) Untergang -=24.0;

    {
        int AufgangMinuten   = (int)(60.0*(Aufgang - (int)Aufgang)+0.5);
        int AufgangStunden   = (int)Aufgang;
        int UntergangMinuten = (int)(60.0*(Untergang - (int)Untergang)+0.5);
        int UntergangStunden = (int)Untergang;

        if (AufgangMinuten>=60.0) { AufgangMinuten-=60.0; AufgangStunden++; }
        else if (AufgangMinuten<0.0)
        {
            AufgangMinuten+=60.0; AufgangStunden--;
            if (AufgangStunden<0.0) AufgangStunden+=24.0;
        }

        if (UntergangMinuten>=60.0) { UntergangMinuten-=60.0; UntergangStunden++; }
        else if (UntergangMinuten<0)
        {
            UntergangMinuten+=60.0; UntergangStunden--;
            if (UntergangStunden<0.0) UntergangStunden+=24.0;
        }

        if ( aufm )
            *aufm=(AufgangStunden*60+AufgangMinuten);
        if ( unterm )
            *unterm=(UntergangStunden*60+UntergangMinuten);
    }
}
#endif
